/*
 * connection, line ctrl etc..
 */

#ifndef LINE_H
#define LINE_H

#include <vector>
#include <algorithm>
#include <queue>
#include <map>

#include "thread++.h"

#include "modem.h"
#include "packet.h"
#include "util.h"
#include "protocol.h"
#include "discovery.h"
#include "compress.h"

using namespace std;

typedef int sequence;

//******************************************************************************************
// plug, connection...

class connection;

class plug {
public:
  virtual bool connect () abstract;
  virtual void disconnect () abstract;
  virtual void data_poll (class itcp_connection *con) abstract;
  virtual void data_push (packet::const_iterator data, int len) abstract;
};

class connection {
  friend class line_control;

protected:
  // for use by line_control
  recursive_mutex lock;
  
  timestamp tx_inhibit;
  void tx_now () { tx_inhibit.clear (); };
  void tx_at (timeout t) { tx_inhibit.stamp (t); };
  
public:
  msn src_msn, dst_msn;
  
  virtual ~connection() { lock.lock (); };
  
  typedef void (connection::*message)();
  
  /* messages a connection can receive */
  virtual void connect () abstract;
  virtual void listen () abstract;
  virtual void stop () abstract;
  virtual void reset () abstract;
  virtual void dump() abstract; // dump state to stdout
  
  /* wire functions */
  virtual packet send_packet () abstract; // get a packet to send
  virtual void receive_packet (const packet &p) abstract; // push a packet received
  
  /* match function */
  virtual bool match (const packet &p) abstract;
};

class itcp_connection : public connection {
protected:
  u8 src_port, dst_port;
  u8 pflags;
  
  enum {S_CLOSED, S_LISTEN, S_SEND_SYN, S_ESTABLISHED} state, init_state;
  ie_mask mask_tx;
  int frame_tx; // in *bytes*
  queue<u8> ctrl;
  
  // statistics
  long bytes_in, bytes_out;
  long packets_in, packets_out;
  
  // compression
  compressor *comp;
  decompressor *decomp;
  
public:
  queue<u8> data;
  
  mutex lock;
  
  typedef map<sequence,packet> packet_map;
  packet_map p_recv, p_send;
  sequence seq_recv, seq_send, seq_ackn;
  
  exponential_timeout<9000, 4, 3, 40000> retransmit_time;
  bool do_ack; // force ack if necessary
  
  plug &pl; // data source/sink
  
private:
  void push_ctrl (u8 ctrlbyte);
  void push_packet (const packet &p); // push in sequence
  void acknowledge (sequence ack);
public:
  itcp_connection (const char *src, u8 sport, const char *dst, u8 dport, plug &p);
  ~itcp_connection ();
  
  void connect ();
  void listen ();
  void stop ();
  void reset ();
  void dump(); // dump state to stdout
  
  packet send_packet (); // get a packet to send
  void receive_packet (const packet &p); // push a packet received
  bool match (const packet &p);
  void wire_init ();
  void wire_ready ();
};

struct tty_plug : plug {
  int master;
  int pid;
  char ttyname[40];
  char command[200];
  
  tty_plug (const char *cmd)
    {
      pid = 0;
      strcpy (command, cmd);
    };
  
  bool connect ();
  void disconnect ();
  void data_poll (itcp_connection *con);
  void data_push (packet::const_iterator data, int len);
};

extern class line_control : public runnable {
  friend void receiver(void *);
  friend void transmitter(void *);
  
  static recursive_mutex cons_lock;
  static bool cons_changed;
  typedef vector<connection *> convec;
  static convec cons;
  static pthread_t transmitter_tid, receiver_tid;
  
  void *transmitter(void *);
  void *receiver(void *);
  
public:
  void add_connection (connection *con);
  void del_connection (connection *con);
  
  void add_modemdev (const char *path);
  void set_ctrldev (const char *path);
  void command_all (connection::message msg, const char *dst_msn);
  ~line_control ();
} line_control;

#endif

