/*
 * util.C
 * (c) pcg@goof.com
 */

#include <stdio.h>
#include <signal.h>
#include <unistd.h>

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif /* HAVE_SYS_SELECT_H */

#ifdef STDC_HEADERS
#include <string.h> /* for bzero on BSD systems */
#endif

#include "config.h"
#include "util.h"

int log_mask = L_WARN | L_ERROR | L_FATAL;

timeout timestamp::age () const
{
  timeval tv;
  
  gettimeofday (&tv, 0);
  return (tv.tv_sec - tv_sec) * 1000 + (tv.tv_usec - tv_usec) / 1000;
}

#define ON(l) (log_mask & (l) ? "on" : "off")

void parse_logopt(const char *opt)
{
  static const char *lognames[] = {
    "junk", "isdnctrl" , "itcp" , "discovery",
    "pkt", "line", "debug", "info",
    "warn", "error", "fatal", 0
  };
  const char **name = lognames;
  
  if (!opt)
    {
      for (int mask = 1; *name; mask<<=1, name++)
        printf ("log %-14s: %s\n", *name, ((log_mask & mask) ? "on" : "off"));
    }
  else
    {
     int ormask, mask;
     
      for (mask = 1; *name; mask<<=1, name++)
        if (!strncmp (*name, opt+1, strlen (opt+1)))
          break;
      
      if(opt[0] == '-')
        ormask = 0;
      else if (opt[0] == '+')
        ormask = mask;
      else
        {
          printf ("use + to switch on an option, - to switch it off\n");
          return;
        }
      
      if (*name)
        {
          log_mask = log_mask & ~mask | ormask;
          printf ("log %-14s: %s\n", *name, ((log_mask & ormask) ? "on" : "off"));
        }
      else
        printf ("unknown options '%s'\n", opt+1);
    }
}

/****************************************************************************************/
/* debugging.  */

static char **program_args;

static void
stack_trace ()
{
  pid_t pid;
  int in_fd[2];
  int out_fd[2];
  fd_set fdset;
  fd_set readset;
  struct timeval tv;
  int sel, index, state;
  char buffer[256];
  char c;
  static int stack_trace_done;
  char *args[4];
  
  sprintf (buffer, "%d", (int) getpid ());

  args[0] = "gdb";
  args[1] = program_args[0];
  args[2] = buffer;
  args[3] = 0;
  
  printf ("%s %s %s\n", args[0], args[1], args[2]);
  sleep (86400);
  
  stack_trace_done = 0;
//  signal (SIGCHLD, stack_trace_sigchld);

  if ((pipe (in_fd) == -1) || (pipe (out_fd) == -1))
    {
      perror ("could open pipe");
      _exit (0);
    }

  pid = fork ();
  if (pid == 0)
    {
      close (0); dup (in_fd[0]);   /* set the stdin to the in pipe */
      close (1); dup (out_fd[1]);  /* set the stdout to the out pipe */
      close (2); dup (out_fd[1]);  /* set the stderr to the out pipe */

      execvp (args[0], args);      /* exec gdb */
      perror ("exec failed");
      _exit (0);
    }
  else if (pid == (pid_t) -1)
    {
      perror ("could not fork");
      _exit (0);
    }

  FD_ZERO (&fdset);
  FD_SET (out_fd[0], &fdset);

  write (in_fd[1], "backtrace\n", 10);
  write (in_fd[1], "p x = 0\n", 8);
  write (in_fd[1], "quit\n", 5);

  index = 0;
  state = 0;

  while (1)
    {
      readset = fdset;
      tv.tv_sec = 1;
      tv.tv_usec = 0;

      sel = select (FD_SETSIZE, &readset, NULL, NULL, &tv);
      if (sel == -1)
        break;

      if ((sel > 0) && (FD_ISSET (out_fd[0], &readset)))
        {
          if (read (out_fd[0], &c, 1))
            {
              switch (state)
                {
                case 0:
                  if (c == '#')
                    {
                      state = 1;
                      index = 0;
                      buffer[index++] = c;
                    }
                  break;
                case 1:
                  buffer[index++] = c;
                  if ((c == '\n') || (c == '\r'))
                    {
                      buffer[index] = 0;
                      fprintf (stdout, "%s", buffer);
                      state = 0;
                      index = 0;
                    }
                  break;
                default:
                  break;
                }
            }
        }
      else if (stack_trace_done)
        break;
    }

  close (in_fd[0]);
  close (in_fd[1]);
  close (out_fd[0]);
  close (out_fd[1]);
  sleep (86400);
  _exit (0);
}

static void
debug_signal (int sig_num)
{
  static int caught_fatal_sig = 0;
  char *sig;

  if (caught_fatal_sig)
    kill (getpid (), sig_num);
  caught_fatal_sig = 1;

  switch (sig_num)
    {
    case SIGHUP:
      sig = "sighup";
      break;
    case SIGINT:
      sig = "sigint";
      break;
    case SIGQUIT:
      sig = "sigquit";
      break;
    case SIGBUS:
      sig = "sigbus";
      break;
    case SIGSEGV:
      sig = "sigsegv";
      break;
    case SIGPIPE:
      sig = "sigpipe";
      break;
    case SIGTERM:
      sig = "sigterm";
      break;
    default:
      sig = "unknown signal";
      break;
    }

  printf ("\n** ERROR **: %s caught\n", sig);
  stack_trace ();
}


void init_debug (char **args)
{
  program_args = args;
  printf ("debugging, setting signal handlers\n");
//  signal (SIGHUP, debug_signal);
//  signal (SIGINT, debug_signal);
//  signal (SIGQUIT, debug_signal);
  signal (SIGBUS, debug_signal);
  signal (SIGSEGV, debug_signal);
  signal (SIGPIPE, debug_signal);
  signal (SIGTERM, debug_signal);
}

